
open Ast
(*change a file name from "abc" to 'abc' *)
let remove_qoute str =
	let len = String.length str in
	let _= String.set str 0 '\'' in
	let _= String.set str (len-1) '\'' in
	str
	
let build_binary_operator func_file op =
	match op with
		|Plus -> let _= output_string func_file " + " in func_file
		|Minus-> let _= output_string func_file " - " in func_file
		|Times-> let _= output_string func_file " * " in func_file
		|Divide-> let _= output_string func_file " / " in func_file
		|Mod->  func_file (*because mod in matlab is a funciton, generate the code sepeartely in build_expr *)
		|Lessthan-> let _= output_string func_file " < " in func_file
		|Greaterthan-> let _= output_string func_file " > " in func_file
		|Lessequal-> let _= output_string func_file " <= " in func_file
		|Greaterequal-> let _= output_string func_file " >= " in func_file
		|Equal-> let _= output_string func_file " == " in func_file
		|Notequal-> let _= output_string func_file " ~= " in func_file
		|Logicaland-> let _= output_string func_file " && " in func_file
		|Logicalor-> let _= output_string func_file " || " in func_file
 
let build_unary_operator func_file op = 
	match op with 
	|Negation ->let _= output_string func_file " ~ " in func_file
  |Ninus ->let _= output_string func_file " - " in func_file
	
let build_id func_file identifier =
	let i = identifier in
	let _ = output_string func_file i in
	func_file
	
let build_const func_file const =
	match const with 
		|Intc(v) -> let _= output_string func_file (string_of_int v) in func_file  
		|Charc(v) -> let _= output_string func_file "'" in
		             let _= output_string func_file (String.make 1 v) in 
								 let _= output_string func_file "'" in
								 func_file 
		|Floatc(v) -> let _= output_string func_file (string_of_float v) in func_file 
			
let build_init func_file init =
	match init with
		| Init (var_name,var_val) -> match var_val with
																	|None -> func_file
																	|Consopt(x) -> let _= build_id func_file var_name in
																								 let _= output_string func_file "=" in
																								 let _= build_const func_file x in 
																								 let _= output_string func_file ";\n" in
																								 func_file	


let rec build_expr func_file expr =
	match expr with
		|Lval (lval)-> build_lvalue func_file lval
		|Const(const)->( match const with
										| None -> func_file
										| Consopt(c) -> build_const func_file c )
		|Assign(lval,e)->   let _ = build_lvalue func_file lval in
												let _= output_string func_file "=" in
												build_expr func_file e
		|Binop (e1,op,e2)->let _= output_string func_file "(" in
											 let _= 
												( match op with
												| Mod -> let _= output_string func_file "mod(" in
												         let _= build_expr func_file e1 in
																 let _= output_string func_file "," in
																 let _= build_expr func_file e2 in
																 let _= output_string func_file ")" in
																 func_file
											  | _ -> 
													     let _= build_expr func_file e1 in
															 let _= build_binary_operator func_file op in														
												       build_expr func_file e2) in
											let _= output_string func_file ")" in
											func_file
															
		|Unaryop (op,e1)->  let _= build_unary_operator func_file op in
												let _= output_string func_file "(" in
												let _ = build_expr func_file e1 in
												let _= output_string func_file ")" in
												func_file
		|Comma (e1,e2)-> let _=build_expr func_file e1 in
										 let _= output_string func_file "," in
										 build_expr func_file e2
		|This (e1,e2)-> let _= output_string func_file "this(" in
										let _=build_expr func_file e1 in
										let _= output_string func_file "," in
										let _=build_expr func_file e2 in
										let _= output_string func_file ",1)" in
										func_file
		|Funcall (f_name,e_list)-> let _=build_id func_file f_name in
															 let _= output_string func_file "(" in
															 (* helper function to print arguments in a function call *)
															 let build_expr2 func_file e = 
																output_string func_file ",";  
																build_expr func_file e 
															 in
															(* if there is 0 argument passed, "_" case will be match so nothing printed *)
															(* if there is 1 argument passed, "hd::tl" will be matched and the first expressino is printed normally *)
															(*                                 "tl" is [], so List.fold_left will do nothing but return func_file *)
															(* if there is more than 1 arguments, "hd::tl" matched and "hd" printed normally and before print each *)
															(*                                    expression in "tl", an extra "," will be added*) 
															 let _=(match e_list with
																| hd::tl -> let _=build_expr func_file hd in
																            List.fold_left build_expr2 func_file tl 
																| _ -> func_file) in
															 let _= output_string func_file ")" in
															 func_file
and build_lvalue func_file lval =
	match lval with
		|Variable (var) -> build_id func_file var
		|Array(var,e) -> let _= build_id func_file var in
										 let _=output_string func_file "(" in
										 let _= build_expr func_file e in
										 let _=output_string func_file ")" in
										 func_file
		|Thisl(e) -> let _= build_expr func_file e in func_file (*(match e with 
									|Comma(e1,e2) -> let _= output_string func_file "this(" in
																	 let _=build_expr func_file e1 in
																	 let _= output_string func_file "," in
																	 let _=build_expr func_file e2 in
																	 let _= output_string func_file ")" in
																	 func_file 
									| _-> func_file )*)
(* get all identifiers from video declaration so that we can initialize it later, which is required by matlab lib*)
let get_id ids init = 
	match init with
		| Init (var_name,_) -> var_name::ids

let build_video func_file identifier =
	let _=output_string func_file (identifier^".cdata=[];\n"^identifier^".colormap=[];\n") in
	func_file
	
let build_decl func_file decl =
	match decl with
		|  VarDecl(var_decl) -> (match var_decl with
									           | Ddecl(t,init_list) -> (match t with 
															                         | Video -> let _ =List.fold_left build_init func_file init_list in
																																		let all_ids= List.fold_left get_id [] init_list in 
																																		List.fold_left build_video func_file all_ids 
																																		
																											 | _ -> List.fold_left build_init func_file init_list
																											)
														)
														
    |  ArraDecl(_) -> func_file
	
let rec build_statement func_file stmt =
	 match stmt with 
		| Expression(expr) -> let _=build_expr func_file expr in
													let _= output_string func_file ";\n" in
													func_file
		| Compound(decls,st) -> let _= List.fold_left build_decl func_file decls in
													 let _=List.fold_left build_statement func_file st in
													 let _= output_string func_file "\n" in
													 func_file
	  | If(expr,st1,st2) -> let _= output_string func_file "if " in (*FIXME: ELSE*)
			                    let _= build_expr func_file expr in
													let _= output_string func_file "\n" in									
													let _ = build_statement func_file st1 in
													let _= output_string func_file "else \n" in
													let _ = build_statement func_file st2 in
													let _= output_string func_file "\nend\n" in
													func_file
		| While(expr,st) ->   let _= output_string func_file "while " in 
			                    let _= build_expr func_file expr in
													let _= output_string func_file "\n" in
													let _ = build_statement func_file st in
													let _= output_string func_file "\nend \n" in
													func_file
		| For(expr1,expr2,expr3,st)  ->  let _= output_string func_file "for " in 
			                               let _= build_expr func_file expr1 in
																		 let _=  match expr3 with 
																			| Assign(_,sub_e1) -> (output_string func_file ":" ;
																			                       match sub_e1 with (* for (i=1;i<10;i=i+1) is the only form that is allowed *)
																														 | Binop(_,_,e2) -> build_expr func_file e2
																														 | _ -> func_file )(*FIXME*)
																		 	| _ -> func_file	in								  
																		 let _= output_string func_file ":" in 
																		 let _= match expr2 with
																			| Binop(_,_,e) -> build_expr func_file e 
																			| _ -> func_file in
																	   let _= output_string func_file "\n" in
																		 let _= build_statement func_file st in
																		 let _= output_string func_file "\nend \n" in 
																		 func_file
																		 
		| Continue-> let _= output_string func_file "continue \n" in func_file 
		| Break -> let _= output_string func_file "break; \n" in func_file   
		| Return(expr) -> let _= output_string func_file "ret_vopl_xxx=" in
											let _= build_expr func_file expr in
											let _= output_string func_file ";\n return \n" in
											func_file
		| Load(var,file,e1,e2) -> let file = remove_qoute file in
															let _ = build_id func_file var in
															let _= output_string func_file ("=load_video("^file^",") in
															let _= build_expr func_file e1 in
															let _= output_string func_file "," in
															let _= build_expr func_file e2 in
															let _= output_string func_file ");\n" in
															func_file
		| Store(var,file) -> let file = remove_qoute file in
												 let _= output_string func_file "store_video(" in
												 let _= build_id func_file var in
												 let _= output_string func_file (","^file^");\n")
												 in func_file
		| Insert(var1,var2,e) -> 
			                       let _ = build_id func_file var2 in
														 let _= output_string func_file "=insert_video(" in
														 let _ = build_id func_file var2 in
														 let _= output_string func_file "," in
														 let _ = build_id func_file var1 in
														 let _= output_string func_file "," in
														 let _= build_expr func_file e in
														 let _= output_string func_file ");\n" in
														 func_file
		| Delete(var,e1,e2) ->   
														 let _ = build_id func_file var in
														 let _= output_string func_file "=delete_video(" in
														 let _ = build_id func_file var in
														 let _= output_string func_file "," in
														 let _= build_expr func_file e1 in
														 let _= output_string func_file "," in
														 let _= build_expr func_file e2 in
														 let _= output_string func_file ");\n" in
														 func_file
		| Copy(var1,e1,e2,var2) ->    
			                           let _ = build_id func_file var2 in
														 		 let _= output_string func_file "=copy_video(" in
														     let _ = build_id func_file var1 in
														     let _= output_string func_file "," in													
														     let _= build_expr func_file e1 in
														     let _= output_string func_file "," in
														     let _= build_expr func_file e2 in
																 let _= output_string func_file "," in
															   let _ = build_id func_file var2 in
														     
														     let _= output_string func_file ");\n" in
														     func_file
		(* update v from index1 to index2 { ....} *)
		(* ==>                                    *)
		(* for v_index = index1:index2            *)
		(*   this=v(:,:,v_index);                 *)  
		(*    .......                             *)
		(* end                                    *)
	 
		| Update(var,e1,e2,st) -> 
			                        let _= output_string func_file "for v_index = " in
															let _ = build_expr func_file e1 in
															let _= output_string func_file ":" in
															let _= build_expr func_file e2 in
															let _= output_string func_file "\nthis=" in
															let _= build_id func_file var in
															let _= output_string func_file "(v_index).cdata;\n" in
															let _= build_statement func_file st in
															let _= output_string func_file "\nend \n" in                                      
	 														func_file



let build_function_type func_file t =
	match t with 
		|Int
		|Char
		|Float
		|Video -> let _ =output_string func_file " ret_vopl_xxx=" in
		           func_file
		|Void  -> let _ =output_string func_file " " in
		           func_file

let build_arg func_file para =
	match para with
		| Arg(_, para_name) ->
	let func_file = build_id func_file para_name in
	func_file 
	

let build_function func_file f =
	match f with
		| Func(t, f_name, args, stmts) ->
	let _= output_string func_file "function" in
	let func_file = build_function_type func_file  t in
	let func_file = build_id func_file f_name in
	let _ = output_string func_file "(" in
	(*let _ = print_int (List.length args) in*)
	(* helper function to print arguments in a function def *)
	let build_arg2 func_file ar = 
			output_string func_file ",";
			build_arg func_file ar 
	in 
	let _=(match args with
			| hd::tl -> let _=build_arg func_file hd in
											List.fold_left build_arg2 func_file tl 
			| _ -> func_file) in
	let _= output_string func_file ")\n" in
	let func_file = build_statement func_file stmts in
	func_file


let build p filename=
	match p with
		| Pro(prog) ->
	let n = String.index filename '.' in
	let filename = String.sub filename 0 n in 
	let function_list =List.rev prog in
	let main_channel =
		open_out ( (filename)^".m") in
  List.fold_left build_function main_channel function_list
